from .ei import Normed, empirical_init, get_wrapped_submodules, wrap_all_leaf_modules
from .block_sparse import HyperCubies, hypercubies
from .grad_fn_autotest import autotest_custom_grad_fn

# define public-facing objects:
__all__ = ["Normed", "empirical_init", "get_wrapped_submodules", "wrap_all_leaf_modules", "HyperCubies", "hypercubies", "autotest_custom_grad_fn"]



import torch


class DummyCtx:
  def save_for_backward(*args):
    pass

def autotest_custom_grad_fn(grad_fn, input_shapes, output_shape, verbose=False, epsilon=0.01):
  """ Automatically test the correctness of custom (user-defined) differentiable autograd functions.
      Compares the results from using user-defined backward to the results given by autograd.
      If your implementation would not work with autograd, then this test is not representative.
      grad_fn: is the class where you defined the forward and backward methods
        (so grad_fn.apply would be the actual differentiable func, don't pass that!)
      input_shapes: is a list of tensor shapes expected by your function
        eg. [(1, 2, 7), (5,)]
      output_shape: is the shape of the output your function will produce
        eg. (5, 7) """
  def print_if_verbose(*args):
    if verbose:
      print(*args)
  inputs = [torch.randn(*shape, requires_grad=True) for shape in input_shapes]
  grad = torch.randn(*output_shape)
  # compute gradients using autograd
  print_if_verbose("forward")
  output = grad_fn.forward(DummyCtx(), *inputs)
  print_if_verbose("automatic backward")
  output.backward(grad)
  print_if_verbose("done.")
  expected_grads = []
  for input_tensor in inputs:
    expected_grads.append(input_tensor.grad)
    input_tensor.grad = None
  # compute gradients using user-defined code
  print_if_verbose("forward")
  output = grad_fn.apply(*inputs)
  print_if_verbose("user-defined backward")
  output.backward(grad)
  print_if_verbose("done.")
  actual_grads = []
  for input_tensor in inputs:
    actual_grads.append(input_tensor.grad)
    input_tensor.grad = None
  for actual_grad, expected_grad in zip(actual_grads, expected_grads):
    rms_err = torch.sqrt(((actual_grad - expected_grad)**2).sum())
    print_if_verbose("RMS Error: ", rms_err)
    assert rms_err < epsilon, f"error is above threshold of {epsilon}"




import torch
import torch.nn as nn


# define which things will be exported
__all__ = ["Normed", "empirical_init", "get_wrapped_submodules", "wrap_all_leaf_modules"]


# constants
EPSILON = 1e-12


def names_bias_param(s):
  s = s.casefold()
  return "bias" in s


class GradScale(torch.autograd.Function):
  """ identity function, scales the backwards-flowing gradient by a specified amount """
  @staticmethod
  def forward(ctx, x, a):
    ctx.a = a
    return x
  @staticmethod
  def backward(ctx, grad_y):
    return ctx.a*grad_y, None
grad_scale = GradScale.apply



def bigO(tensor):
  with torch.no_grad():
    if torch.is_complex(tensor):
      return float(torch.sqrt(torch.real(tensor*torch.conj(tensor)).mean()))
    else:
      return float(torch.sqrt((tensor**2).mean()))

def inplace_add(t, x):
  """ add x to tensor t, leaving t in place """
  t.add_(x + torch.zeros_like(t))
  return t


class Normed(nn.Module):
  """ wraps a module and provides proper normalization on its inputs and outputs """
  def __init__(self, module):
    super().__init__()
    self.module = module
    # new weights will be randn(), correct for this by measuring current weights since they're approximately correct
    parameter_scale = min(bigO(param) for param in self.module.parameters() if bigO(param) > 0)
    # register all these values as buffers so they will be included in the state dict:
    self.register_buffer("scale_actv_y", torch.tensor(parameter_scale))
    self.register_buffer("scale_grad_x", torch.tensor(parameter_scale))
    self.register_buffer("scale_grad_y", torch.tensor(1.0))
    self.init_all_parameters()
  def forward(self, *inputs):
    scaled_inputs = [
      grad_scale(x, self.scale_grad_x)
      for x in inputs ]
    y = self.module(*scaled_inputs)
    return grad_scale(self.scale_actv_y*y, self.scale_grad_y)
  def init_all_parameters(self):
    with torch.no_grad():
      for nm, param in self.module.named_parameters():
        if names_bias_param(nm):
          nn.init.constant_(param.data, 0.)
        else:
          nn.init.normal_(param.data)


class ModuleMonitor:
  def __init__(self, module):
    assert isinstance(module, Normed), "can't track modules unless they're wrapped in a normalization module"
    self.module = module
  def zero_grad(self):
    self.module.zero_grad()
  def reset(self):
    self.module.init_all_parameters()
    self.zero_grad()
  def set_activation_hook(self):
    def hook(mod, x, y):
      self.input_activations = [xi.detach() for xi in x]
      self.activation = y.detach()
    self.activation_hook_handle = self.module.register_forward_hook(hook)
  def clear_activation_hook(self):
    self.activation_hook_handle.remove()
  def set_gradient_hook(self):
    def hook(mod, grad_x, grad_y):
      assert len(grad_y) == 1
      self.input_gradient = grad_y[0].detach()
      self.gradient = [
        (grad_xi if grad_xi is not None else None)
        for grad_xi in grad_x ]
    self.gradient_hook_handle = self.module.register_full_backward_hook(hook)
  def clear_gradient_hook(self):
    self.gradient_hook_handle.remove()
  def set_eval_order_hook(self, eval_order_list):
    self.eval_count = 0
    def hook(mod, x, y):
      eval_order_list.append(self)
      self.eval_count += 1
    self.eval_order_hook_handle = self.module.register_forward_hook(hook)
  def clear_eval_order_hook(self):
    assert self.eval_count > 0, "module was never called"
    assert self.eval_count == 1, "module was called too many times: %d" % self.eval_count
    self.eval_order_hook_handle.remove()
  def collect_param_update_data(self, test_lr):
    self.param_gradients = {nm: param.grad for nm, param in self.module.named_parameters()}
    with torch.no_grad():
      prev_activation = self.activation.clone()
      for nm, param in self.module.named_parameters():
        # perform a small update to this particular param in the direction of the gradient
        grad = self.param_gradients[nm]
        inplace_add(param, -test_lr*grad)
      # rerun the module with all parameters updated
      self.activation = self.module(*self.input_activations)
      # save the result, corrected for the scaling we did
      self.activation_update = (self.activation - prev_activation)/test_lr
  def show(self):
    def tensor_to_datastr(t):
      bigO_t = bigO(t)
      if torch.is_complex(t):
        t = torch.real(t)
      return "O(%f), z=%d%s" % (bigO_t, round(100*float(t.mean()/(EPSILON + bigO_t))), "%")
    print(type(self.module.module))
    print("X ---(%f)--> [...] ---(%f)--> Y" % (1.0, self.module.scale_actv_y))
    print("X <--(%f)--- [...] <--(%f)--- Y" % (self.module.scale_grad_x, self.module.scale_grad_y))
    print("ACTIVATION")
    print("    %s" % tensor_to_datastr(self.activation))
    print("GRADIENTS")
    for grad in self.gradient:
      if grad is None:
        print("    None")
      else:
        print("    %s" % tensor_to_datastr(grad))
    print("CHANGE IN ACTIVATION")
    print("    %s" % tensor_to_datastr(self.activation_update))
    print("PARAM GRADIENTS")
    for nm in self.param_gradients:
      print("    %s: %s" % (nm, tensor_to_datastr(self.param_gradients[nm])))
    print()


def get_eval_order(monitors, run_model):
  """ Run the model once to determine evaluation order """
  eval_order = []
  for monitor in monitors:
    monitor.set_eval_order_hook(eval_order)
  run_model() # SIDE EFFECT: eval_order should now be populated
  for monitor in monitors:
    monitor.clear_eval_order_hook()
  return eval_order

def tune_outgoing_act(monitors, run_model):
  """ Tune Outgoing Activations: Fine tune by running a full forward pass through the model several times.
      For stability reasons, it's best to tune the earliest-evaluated modules first.
      Expects monitors sorted from earliest-evaluated to latest. """
  for monitor in monitors:
    monitor.set_activation_hook()
  for monitor in monitors:
    for i in range(0, 5):
      run_model()
      bigO_actv = bigO(monitor.activation)
      # square root the correction for better stability
      monitor.module.scale_actv_y /= EPSILON + bigO_actv**0.5
  for monitor in monitors:
    monitor.clear_activation_hook()

def tune_outgoing_grad(monitors, run_model):
  """ Tune Outgoing Gradients: Fine tune by running a full backward pass through the model several times.
      For stability reasons, it's best to tune the last-evaluated modules first.
      Expects monitors sorted from earliest-evaluated to latest. """
  for monitor in monitors:
    monitor.set_gradient_hook()
  for monitor in reversed(monitors):
    for i in range(0, 5):
      run_model()
      bigO_grad = 0.
      for grad in monitor.gradient:
        if grad is not None:
          bigO_grad = max(bigO(grad), bigO_grad)
      if bigO_grad > EPSILON:
        # square root the correction for better stability
        monitor.module.scale_grad_x /= EPSILON + bigO_grad**0.5
  for monitor in monitors:
    monitor.clear_gradient_hook()

def tune_incoming_grad(monitors, run_model, test_lr):
  """ Tune Incoming Gradients:
      Fine tune by running a full forward and backward pass through the model several times.
      At this point, the product of the scalings for grad_x and grad_y should not change. """
  for monitor in monitors:
    monitor.set_activation_hook()
    monitor.set_gradient_hook()
  for i in range(0, 20):
    run_model()
    # changes are internal to module, so we can update all modules on each run
    for monitor in monitors:
      monitor.collect_param_update_data(test_lr)
      bigO_param_update = bigO(monitor.activation_update)
      # square root to get correction, then square root that for better stability
      scaling_factor = EPSILON + bigO_param_update**0.25
      monitor.module.scale_grad_x *= scaling_factor
      monitor.module.scale_grad_y /= scaling_factor
  for monitor in monitors:
    monitor.clear_activation_hook()
    monitor.clear_gradient_hook()

def show_tuning_results(monitors, run_model, test_lr):
  for monitor in monitors:
    monitor.set_activation_hook()
    monitor.set_gradient_hook()
  run_model()
  print("\nTuning Results:")
  for monitor in monitors:
    monitor.collect_param_update_data(test_lr)
    monitor.show()
  for monitor in monitors:
    monitor.clear_activation_hook()
    monitor.clear_gradient_hook()



def empirical_init(wrappers, forward, get_input, get_loss, batchsz=64, test_lr=1e-5):
  """ empirically initialize a model, requires that model work on batched data
      wrappers  - list of empirical_init wrappers to tune, can be obtained by calling .get_tunable_modules()
      forward   - function to call to evaluate the model being tuned
      get_input - function to call to get a random representative input for the model, takes batchsz as parameter
      get_loss  - function to call to get a random representative loss for the model, takes model output as parameter
      batchsz   - size of the batch, this is what we do a statistical average over
      test_lr   - test learning rate for when we try adjusting the parameters """
  monitors = [ModuleMonitor(wrapper) for wrapper in wrappers]
  
  def run_model():
    """ Re-initializes the model, runs the model, and computes gradients. """
    for monitor in monitors:
      monitor.reset()
    dummy_input = get_input(batchsz)
    model_output = forward(dummy_input)
    dummy_loss = get_loss(model_output)
    dummy_loss.backward()
  
  monitors = get_eval_order(monitors, run_model)
  tune_outgoing_act(monitors, run_model)
  tune_outgoing_grad(monitors, run_model)
  tune_incoming_grad(monitors, run_model, test_lr)
  show_tuning_results(monitors, run_model, test_lr)
  # finally, we reinitialize all the modules a final time
  for monitor in monitors:
    monitor.reset()


def get_wrapped_submodules(module):
  return [mod for mod in module.modules() if isinstance(mod, Normed)]

def is_wrappable_leaf_module(module):
  submodules = sum(1 for _ in module.modules())
  parameters = sum(1 for _ in module.parameters())
  return submodules == 1 and parameters > 0

def wrap_all_leaf_modules(module_init_fn):
  def new_module_init_fn(self, *args, **kwargs):
    ans = module_init_fn(self, *args, **kwargs)
    modules = {}
    wrappable_leaf_modules = set()
    for nm, module in self.named_modules():
      modules[nm] = module
      if is_wrappable_leaf_module(module):
        print("Wrapping %s" % nm)
        wrappable_leaf_modules.add(nm)
    for module_nm in wrappable_leaf_modules:
      if module_nm != "":
        parent_module = modules[".".join(module_nm.split(".")[:-1])]
        setattr(parent_module, module_nm.split(".")[-1], Normed(modules[module_nm]))
    return ans
  return new_module_init_fn

# --- TODO List ---
# TODO: make things work better when modules have multiple inputs
# TODO: add a "wrap all children" decorator, maybe a "wrap all descendants" one too
# TODO: make things work better when modules have multiple outputs
# TODO: make epsilon a passable parameter rather than a constant





import torch
import torch.nn as nn


class HyperCubiesFunction(torch.autograd.Function):
  """ A block-sparse function whose blocks correspond to vertices of a hypercube.
      Vertices are represented as binary integers (https://blog.plover.com/math/cube-components.html)
      two vertices are connected by an edge when popcount(a ^ b) == 1.
      Storage of weights works by having the first dimension be which vertex to end at,
      second dimension is which direction do we move to get to the starting vertex. Special case:
      dim corresponds to no movement. """
  @staticmethod
  def forward(ctx, W, x):
    """ dimensions of W: (2**dim, dim + 1, chan_out, chan_in), dimensions of x: (batch, 2**dim, chan_in) """
    # check validity of inputs
    device = x.device
    assert W.device == device, "Expected both input tensors to be on the same device"
    W_n_cubies, dim_pp, chan_out, chan_in = W.shape
    dim = dim_pp - 1
    n_cubies = 2**dim
    assert W_n_cubies == n_cubies
    batch = x.shape[0]
    assert x.shape[1] == n_cubies
    assert x.shape[2] == chan_in
    # perform actual forwards op
    ctx.save_for_backward(W, x)
    ans = torch.zeros((batch, n_cubies, chan_out), device=device)
    for v_out in range(n_cubies):
      for i_b in range(dim + 1):
        v_in = (v_out ^ (1<<i_b)) % n_cubies # vertex in direction i_b from v
        ans[:, v_out] += (W[v_out, i_b] @ (x[:, v_in].reshape(batch, chan_in, 1))).reshape(batch, chan_out)
    return ans
  @staticmethod
  def backward(ctx, grad):
    W, x = ctx.saved_tensors
    device = W.device
    batch = x.shape[0]
    dim = W.shape[1] - 1
    chan_out, chan_in = W.shape[2:]
    n_cubies = 2**dim
    # compute gradient of W
    grad_W = torch.zeros(W.shape, device=device)
    grad_x = torch.zeros(x.shape, device=device)
    for v_out in range(n_cubies):
      for i_b in range(dim + 1):
        v_in = (v_out ^ (1<<i_b)) % n_cubies # vertex in direction i_b from v
        grad_W[v_out, i_b] = (grad[:, v_out].reshape(batch, chan_out, 1) * x[:, v_in].reshape(batch, 1, chan_in)).sum(0)
        grad_x[:, v_in] += (grad[:, v_out].reshape(batch, 1, chan_out) @ W[v_out, i_b]).reshape(batch, chan_in)
    return grad_W, grad_x


hypercubies = HyperCubiesFunction.apply


class HyperCubies(nn.Module):
  """ A linear layer, but block sparse according to the hyper-cubies scheme """
  def __init__(self, chan_in, chan_out, dim):
    super().__init__()
    norm_factor = ((dim + 1)*max(chan_in, chan_out))**(-0.5)
    self.W = nn.Parameter(norm_factor*torch.randn(2**dim, dim + 1, chan_out, chan_in))
    self.dim = dim
  def forward(self, x):
    return hypercubies(self.W, x)
  def density_ratio(self):
    return (self.dim + 1)/2**self.dim



from setuptools import setup

required_packages = [
  "torch",
]

with open("README.md", "r") as f:
  long_description = f.read()

setup(
  name="empirical_init",
  version="0.0.2",
  description="Automatically initialize weights in pytorch modules using a hacky empirical method.",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/pb1729/empirical-init",
  author="Phillip Bement",
  author_email="{author_first_name}{author_last_initial}@fastmail.com",
  license="BSD 3-clause",
  packages=["empirical_init"],
  install_requires=required_packages,
)




import torch
import torch.nn as nn
from empirical_init import wrap_all_leaf_modules, get_wrapped_submodules, empirical_init

# Test Code:
def test_empirical_init():
  import torch.nn.functional as F

  class SubmodulesModule(nn.Module):
    def __init__(self, W, device="cpu"):
      super().__init__()
      self.lin0 = nn.Linear(W, W, bias=False, device=device)
      self.lin1 = nn.Linear(W, W, bias=False, device=device)
    def forward(self, x):
      return self.lin1(torch.tanh(self.lin0(x)))
  
  class TestModel(nn.Module):
    @wrap_all_leaf_modules
    def __init__(self, device="cpu"):
      super().__init__()
      self.device = device
      self.lin0 = nn.Linear(20, 64, device=device)
      self.lin1 = nn.Linear(64, 128, device=device)
      self.lin2 = nn.Linear(128, 256, bias=False, device=device)
      self.smdm = SubmodulesModule(256, device=device)
      self.lin3 = nn.Linear(256, 128, bias=False, device=device)
      self.lin4 = nn.Linear(128, 32, bias=False, device=device)
      self.lin5 = nn.Linear(32, 2, bias=False, device=device)
    def forward(self, x):
      logits = self.lin5(
        F.relu(self.lin4(
        F.relu(self.lin3(
        F.relu(self.smdm(
        F.relu(self.lin2(
        F.relu(self.lin1(
        F.relu(self.lin0(
        x)))))))))))))
      return F.softmax(logits, dim=-1)
  def loss(model_output, labels):
    return -torch.log(model_output.gather(-1, labels)).mean()
  def dummy_input(batch):
    """ a method we define so we can use empirical_init() on this model """
    return torch.randn(batch, 20)
  def dummy_loss(model_output):
    """ a method we define so we can use empirical_init() on this model """
    batch, *_ = model_output.shape
    labels = torch.randint(2, (batch, 1))
    return loss(model_output, labels)
  
  model = TestModel()
  empirical_init(
    get_wrapped_submodules(model), model,
    dummy_input, dummy_loss,
    batchsz=64)


if __name__ == "__main__":
  test_empirical_init()


